#!/usr/bin/env python

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Hasher: frontend and backend processing thread classes
'''

import sys
import time
import os.path
import optparse

from PyQt4.QtCore import *
from PyQt4.QtGui import *

from Ui_hashform import Ui_HashForm
from networking import ApiHandler,AES
from cfg import CONFIG,SessionConfig
import ed2k

IS_POSIX = os.name == 'posix'

class FileEntry():
    'Structure storing the status of a single file to be hashed'
    def __init__(self, p):
        self.path = p
        (dir,self.fname) = os.path.split(p)
        # Posix systems need an extra conversion to QString to display the unicode file names nicely in the table
        if IS_POSIX:
            self.fname = QString.fromUtf8(self.fname)
        self.valid = True
        self.status = "Click hash to begin"
        self.ed2k  = ""
        self.crc = ""
        self.crcmatch = ""
        self.size = 0

class FileHashModel(QAbstractTableModel):
    'Model to allow the list of FileEntry objects to be viewed in the QTableView'
    def __init__(self,parent=None):
        QAbstractTableModel.__init__(self,parent)
        self.files = []
    def rowCount(self, index=None):
        return len(self.files)
    def columnCount(self, index=None):
        return 5
    def data(self, index,  role=Qt.DisplayRole):
        if (index.isValid() and role == Qt.DisplayRole):
            f = self.files[index.row()]
            if index.column() == 0:
                return f.status
            elif index.column() == 1:
                return f.fname
            elif index.column() == 2:
                return f.ed2k
            elif index.column() == 3:
                return f.crc
            elif index.column() == 4:
                return f.crcmatch
        return QVariant()
    def headerData(self, section,  orientation,  role=Qt.DisplayRole):
        if (orientation == Qt.Horizontal and role==Qt.DisplayRole):
            if section==0:
                return "Status"
            elif section==1:
                return "Name"
            elif section==2:
                return "Hash"
            elif section==3:
                return "CRC32 Sum"
            elif section==4:
                return "Valid"
        return QVariant()
    def doUpdated(self, i):
        self.dataChanged.emit(self.createIndex(i, 1), self.createIndex(i, 2));
    

class Hasher(QThread):
    'Thread to hash files in the background'
    def __init__(self,a,parent=None):
        QThread.__init__(self, parent)
        self.files = []
        self.options = {}
        self.api = a

    def setFiles(self,fs):
        self.files = fs
        
    def run(self):
        i = 0
        for f in self.files:
            if f.ed2k == '':
                f.status="Waiting"
        for f in self.files:
            if f.ed2k == '':
                f.status="Hashing"
                self.emit(SIGNAL("fileStarted(int)"), i)
                try:
                    v = ed2k.ed2khash(f.path,True)
                    (f.size, f.ed2k, crc) = v
                    f.crc = "%X" % crc
                    if f.crc in str(f.fname).upper():
                        f.crcmatch = "Yes"
                    else:
                        f.crcmatch = ""
                    f.status = 'Hashed'
                    try:
                        print ed2k.ed2klink(f.path, (f.size,f.ed2k))
                    except:
                        pass
                except IOError:
                    f.status = "Unable to Open"
            self.emit(SIGNAL("fileHashed(int)"), i)
            if self.options['action'] != 'None' and f.status!='Unable to open':
                self.api.doMylistaddFile(f.size,f.ed2k,self.options['watched'],i,self.options['action']=='Edit')
                f.status = "Hashed, awaiting response"
                # Sent request to API class: on response, emit signal containing i and success/failure status
            i+=1

class HashGui(QWidget):
    'Main gui for the hasher'
    def __init__(self,api,parent=None,tohash=[],standalone=False,watched=False,start=False):
        QWidget.__init__(self,parent)
        self.ui = Ui_HashForm()
        self.ui.setupUi(self)
        self.setAcceptDrops(True)
        self.connect(self.ui.addButton,SIGNAL('clicked()'),self.addFile)
        self.connect(self.ui.addDirButton,SIGNAL('clicked()'),self.addDir)
        self.connect(self.ui.hashButton,SIGNAL('clicked()'),self.hashFiles)
        self.connect(self.ui.configureButton,SIGNAL('clicked()'),self.launchConfig)
        self.connect(self.ui.removeButton,SIGNAL('clicked()'),self.removeFiles)
        self.files = []
        self.api = api
        self.connect(self.api, SIGNAL('mlafResponse(QString,int)'),self.mlafResponse)
        self.connect(self.api, SIGNAL('noConnection(int)'), self.noConnection)
        self.connect(self.api, SIGNAL('authResponse(int)'), self.authResponse)
        self.H = Hasher(self.api)
        self.connect(self.H, SIGNAL("fileStarted(int)"), self.fileStarted)
        self.connect(self.H, SIGNAL("fileHashed(int)"), self.fileHashed)
        self.ui.watchedBox.setChecked(watched)
        self.model = FileHashModel()
        self.model.files = self.files
        # Add the files which were passed on the command line
        for f in tohash:
            self.recursivelyAddFile(f)
        self.ui.tableView.setModel(self.model)
        if len(self.model.files) < len(tohash):
            QMessageBox.warning(self,"Warning","Not all command line arguments corresponded to files. If you are missing files with unicode filenames, try adding them from within this program.")
        self.standalone = standalone
        if standalone:
            self.setAttribute(Qt.WA_QuitOnClose,True)
        else:
            self.setAttribute(Qt.WA_QuitOnClose,False)
        if len(self.files) > 0:
            self.ui.tableView.resizeColumnsToContents()
        if start:
            self.hashFiles()

    def recursivelyAddFile(self,f,s=None,l=None):
        '''Recursively add the file/directory to the list of files to hash
           If non-None values are passed for s and l, the files will be added to those structures instead'''
        if (type(f)==str):
            f = unicode(f,'utf-8')
        if os.path.isfile(f):
            fe = FileEntry(f);
            if (fe.valid):
                if s!=None:
                    if f not in s:
                        s.add(f)
                        l.append(fe)
                else:
                    self.files.append(fe)
        else:
            for (root,dirs,files) in os.walk(f):
                for fname in files:
                    self.recursivelyAddFile((os.path.join(root,fname)),s,l)
                for dname in dirs:
                    self.recursivelyAddFile((os.path.join(root,dname)),s,l)

        
    
    def addFile(self):
        'Add a list of files (chosen by a file picker dialog)'
        filelist = QFileDialog.getOpenFileNames(self,"Select files to hash",CONFIG.directory())
        self.model.beginInsertRows(QModelIndex(), len(self.files), len(self.files)+len(filelist)-1)
        for f in filelist:
            self.recursivelyAddFile(unicode(f.toUtf8(),'utf-8'))
        self.model.endInsertRows()
        self.ui.tableView.resizeColumnsToContents()

    def addDir(self):
        'Add a directory'
        d = QFileDialog.getExistingDirectory(directory=CONFIG.directory())
        fileset = set()
        filelist = []
        self.recursivelyAddFile(str(d.toUtf8()),fileset,filelist)
        self.model.beginInsertRows(QModelIndex(), len(self.files), len(self.files)+len(filelist)-1)
        for fe in filelist:
            self.files.append(fe)
        self.model.endInsertRows()
        self.ui.tableView.resizeColumnsToContents()
        
    def hashFiles(self):
        'Spawn a thread to hash all files'
        if (not self.H.isRunning()) or self.H.isFinished():
            self.H.setFiles(self.files[:])
            self.H.options['watched'] = self.ui.watchedBox.isChecked()
            self.H.options['action'] = self.ui.mylistBox.currentText()
            self.H.start()
        else:
            QMessageBox.warning(self,"Warning","Hashing already in progress; wait until the first process finishes")

    def launchConfig(self):
        'Launch the configuration dialog and reconnect if the user accepted the dialog (pressed "OK")'
        if (CONFIG.showDialog() ==  QDialog.Accepted):
            if (CONFIG.username() == ""):
                QMessageBox.warning(self,"Warning","If you want to connect to the anidb server, you need to enter a valid username/password")
            else:
                self.api.reconnect()

    def removeFiles(self):
        'Remove all files which are selected in the tableview'
        selection = self.ui.tableView.selectedIndexes()
        # This is the selected cells, so it first finds the set of rows containing those cells
        to_remove = set()
        for s in selection:
            to_remove.add(s.row())
        # Remove the rows from start to end
        to_remove = list(to_remove)
        to_remove.sort()
        i = 0
        for row in to_remove:
            r = row-i
            self.model.beginRemoveRows(QModelIndex(), r, r)
            self.files.pop(r)
            self.model.endRemoveRows()
            i += 1
        
    def fileStarted(self, index):
        'When a hashing of a file is started and finished, update the tableview (since the status has changed)'
        self.model.doUpdated(index)
    def fileHashed(self, index):
        self.model.doUpdated(index)
        
    def mlafResponse(self, status, i):
        'When a response is received from anidb indicating the result of an action, update the stored information'
        f = self.H.files[i]
        f.status = unicode(status).title()
        self.model.doUpdated(i)

    def noConnection(self, t):
        QMessageBox.warning(self,"Warning","No response from server within %i seconds. Check that the correct API server is specified and that your network is allowing the connection. The API server might be down at the moment.<br/>Requests will be ignored until a connection is established." % (t) )
        
    def authResponse(self,code):
        if code == 200:
            pass
        elif code==500:
            QMessageBox.warning(self,"Warning","Login failed; You must set your username and password." )
        else:
            QMessageBox.warning(self,"Warning","Auth failed with response code %i (see console for detailed response). Any further requests will probably fail" % (code) )

    def closeEvent(self,evt):
        ' When the dialog is closed, send a logout request and wait for 0.5 seconds'
        if self.standalone:
            self.api.doLogout()
            self.api.processor.quit()
            time.sleep(0.5)
        evt.accept()

    def dragEnterEvent(self,evt):
        'Accept drag events that contain urls (files to add)'
        if evt.mimeData().hasUrls():
            evt.acceptProposedAction()

    def dropEvent(self,evt):
        fileset = set()
        filelist = []
        for u in evt.mimeData().urls():
            if u.scheme()=='file':
                self.recursivelyAddFile(str(u.toLocalFile().toUtf8()),fileset,filelist)
        self.model.beginInsertRows(QModelIndex(), len(self.files), len(self.files)+len(filelist)-1)
        for fe in filelist:
            self.files.append(fe)
        self.model.endInsertRows()
        self.ui.tableView.resizeColumnsToContents()
        #Tell the drag event's source to ignore the action (i.e. don't delete a file that was "moved" to this window)
        evt.setDropAction(Qt.IgnoreAction)
        evt.accept()


if __name__ == '__main__':
    app = QApplication(sys.argv)

    parser = optparse.OptionParser(usage="%prog [options] [files and directories]")
    settingsgroup = optparse.OptionGroup(parser,"Session-only Configuration options","Setting any of these options will prevent the configuration file from being written")
    settingsgroup.add_option('--username','-u', help='AniDB user name')
    settingsgroup.add_option('--password','-p', help='AniDB password')
    settingsgroup.add_option('--hostname','-a',help='Hostname of AniDB API server')
    settingsgroup.add_option('--port','-n',type=int, help='Port of AniDB API server')
    settingsgroup.add_option('--directory','-d',help="Default directory")
    parser.add_option('--watched','-w',action='store_true',default=False,help="Default to marking files watched")
    parser.add_option('--start','-s',action='store_true',default=False,help="Start hashing automatically on launch")
    parser.add_option('--mylist-action','-m',help="Action to perform by default (add, edit, or none--anything else is treated as 'none')")
    parser.add_option_group(settingsgroup)
    opts,args = parser.parse_args()
    session_only = opts.username or opts.password or opts.hostname or opts.port or opts.directory
    if session_only:
        c = SessionConfig()
        CONFIG = c
        if opts.username:
            c._username = opts.username
        if opts.password:
            c._password = opts.password
        if opts.hostname:
            c._hostname = opts.hostname
        if opts.port:
            c._port = opts.port
        if opts.directory:
            c._directory = opts.directory

    if (CONFIG.username() == ""):
        CONFIG.showDialog()
    api = ApiHandler()
    if (CONFIG.username() == ""):
        QMessageBox.warning(None,"Warning","If you want to connect to the anidb server, you need to enter a valid username/password")
    else:
        api.connectSocket()
    w = HashGui(api,tohash=args,standalone=True,watched=opts.watched,start=opts.start)

    if opts.mylist_action:
        action = opts.mylist_action.title()
        if (action=='Add' or action=='Edit'):
            w.ui.mylistBox.setCurrentIndex(w.ui.mylistBox.findText(action))
        else:
            w.ui.mylistBox.setCurrentIndex(w.ui.mylistBox.findText('None'))

    w.show()

    sys.exit(app.exec_())
