#!/usr/bin/env python

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
This file contains the classes which interact directly with the anidb api
"""

from PyQt4 import QtNetwork, QtCore
from PyQt4.QtCore import *
from PyQt4.QtNetwork import *
from PyQt4.QtGui import QMessageBox
import time,re,hashlib

#The pycrypto module doesn't need to be installed unless encryption is used
try:
    import Crypto.Cipher.AES
    AES = Crypto.Cipher.AES
except:
    AES = None

from cfg import CONFIG

class RequestProcessor(QThread):
    def __init__(self, q, ql, s, parent=None):
        QThread.__init__(self, parent)
        self.queue = q
        self.qlock = ql
        self.last = 0
        self.socket = s
        self.plock = QMutex()
        self.cipher = None
    
    def run(self):
        self.exec_()

    def sendMessage(self,m):
        if self.cipher:
            m = self.encryptMessage(m)
        self.socket.write(m)

    def encryptMessage(self,m):
        pad = 16-(len(m) % 16)
        padded_m = m + chr(pad)*pad
        return self.cipher.encrypt(padded_m)

    def added(self):
        self.plock.lock() 
        self.qlock.lock()
        msg = None
        if len(self.queue) > 0:
            msg = self.queue.pop(0)
        self.qlock.unlock()
        if msg:
            # Sleep if another request was performed within 2 seconds
            diff = time.time() - self.last
            if diff < 2:
                time.sleep(2-diff)
                
            print "Sending '%s'" % msg
            self.last = time.time()
            self.sendMessage(msg)
            self.emit(SIGNAL('messageSent()'))
        self.plock.unlock()


class ApiHandler(QtCore.QObject):
    def __init__(self):
        QObject.__init__(self)
        self.socket = QtNetwork.QUdpSocket()
        self.connect(self.socket,SIGNAL('readyRead()'),self.readyRead)
        self.connected = False
        self.session = None
        self.queue = []
        self.qlock = QMutex()
        self.processor = RequestProcessor(self.queue,self.qlock,self.socket,self)
        self.connect(self,SIGNAL('added()'),self.processor.added)
        self.processor.start()
        self.pingTimer = None
        self.cipher = None
        
    def doOutput(self,data):
        self.emit(SIGNAL('readData(QString)'),data)
    
    def checkConnection(self):
        if not self.session:
            self.emit(SIGNAL('noConnection(int)'), self.timeout)
    
    def connectSocket(self):
        'Connect and AUTH with Anidb (do nothing if already connected'
        if not self.connected:
            self.connected=True
            self.lastPort = 4242
            self.socket.bind(self.lastPort)
            host_addrs = QtNetwork.QHostInfo.fromName(CONFIG.hostname()).addresses() ;
            if len(host_addrs) < 1:
                self.connected = False
                QMessageBox.critical(None,"Error","Unable to resolve host '%s'. Check your internet connection or the configuration." % CONFIG.hostname() )
                return
            self.socket.connectToHost(host_addrs[0],CONFIG.port())
            if CONFIG.encryption():
                if (AES):
                    self.socket.write("ENCRYPT user=%s&type=1&tag=enc" % (CONFIG.username()))
                else:
                    self.connected = False
                    QMessageBox.critical(None,"Error","The pycrypto library is not installed. You must either install the library or reconnect with encryption disabled." )
                    return
            else:
                if (self.session):
                    msg = "UPTIME s=%s&tag=up" % (self.session)
                    status = self.processor.sendMessage(msg)
                else:
                    self.doAuth()
            # Warn user if no response within 5 seconds
            self.timeout = 5
            QTimer.singleShot(self.timeout*1000,  self.checkConnection)

    def doAuth(self):
        if self.connected:
            msg ="AUTH user=%s&pass=%s&protover=3&client=pyaniqthasher&clientver=1&nat=1&tag=auth" % (CONFIG.username(),CONFIG.password())
            status = self.processor.sendMessage(msg)

    def reconnect(self):
        self.doLogout()
        self.socket.disconnectFromHost()
        self.connected=False
        self.connectSocket()
            
    def decryptMessage(self,m):
        d = self.cipher.decrypt(m)
        d.strip(d[-1])
        return d
        
    def readyRead(self):
        'Big function which is called whenever a response datagram is received'
        (data,host,port) = self.socket.readDatagram(self.socket.pendingDatagramSize())
        if (self.cipher):
            try:
                data = self.decryptMessage(data)
                print 'Encrypted',
            except:
                QMessageBox.critical(None,"Error","Encryption/Decryption failed. Did you set the correct api key?" )
        self.emit(SIGNAL('readData(QString)'),data)
        print "Response:'%s'" % data.strip()
        parts = data.split()
        code = 42
        try:
            code = int(parts[1])
        except: pass
        # Choose action based on response type and contents
        if (parts[0] == 'up'):
            if code!=208:
                self.doAuth()
        elif (parts[0] == 'enc'):
            if code == 309:
                QMessageBox.critical(None,"Error","Your account doesn't have an API key defined: go into the account tab of your profile to set one." )
            elif code == 394:
                QMessageBox.critical(None,"Error","Your user account doesn't exist." )
            elif code==209:
                salt = parts[2]
                apikey = str(CONFIG.apikey())
                if apikey=="":
                    QMessageBox.critical(None,"Error","You must specify the API key to use encryption. If your account doesn't have an API key yet, go into the account tab of your profile to set one." )
                    return
                key = hashlib.md5(apikey+salt).digest()
                self.cipher = AES.new(key)
                self.processor.cipher = self.cipher
                self.doAuth()
            else:
                QMessageBox.critical(None,"Error","Failed to enable encryption." )
        elif (parts[0] == 'auth'):
            if code==200 or code==201:
                auth = parts[2]
                hostport = parts[3].split(':')
                port = int(hostport[1])
                if port != self.lastPort:
                    #Behind a NAT, so set up the timer to ping every 5 minutes
                    self.lastPort = port
                    self.pingTimer = QTimer(self)
                    self.connect(self.pingTimer, SIGNAL('timeout()'), self.doPing)
                    self.connect(self.processor, SIGNAL('messageSent()'), self.resetTimer)
                    self.resetTimer()
                self.session = auth
            self.emit(SIGNAL('authResponse(int)'), code)
        elif parts[0][0:4]=='mlaf':
            i = int(parts[0][4:])
            status = data.split(None,2)[-1].split('\n')[0]
            self.emit(SIGNAL('mlafResponse(QString,int)'),  status, i)
        elif len(parts)>2 and parts[1]=='PONG':
            curPort = int(parts[2])
            if curPort != self.lastPort:
                self.doAuth()

    def enqueue(self,msg):
        'Enqueue a request in a thread-safe manner'
        self.qlock.lock()
        self.queue.append(msg)
        self.qlock.unlock()
        self.emit(SIGNAL('added()'))
        
    def resetTimer(self):
        if (self.pingTimer):
            self.pingTimer.start(5 * 60 * 1000)

    def queueLength(self,timeout=-1):
        ret = 0
        locked = self.qlock.tryLock(timeout)
        ret = len(self.queue)
        if locked:
            self.qlock.unlock()
        return ret

    def doLogout(self):
        "Logout immediately (don't bother to queue request)"
        if self.session:
            self.processor.sendMessage("LOGOUT s=%s&tag=bye" % self.session)
            print "Sent logout request"
    
    def doPing(self):
        if self.session:
            self.enqueue("PING nat=1")

    def doMylistaddFile(self,size,hash,watched=False,i=0,edit=False):
        msg = "MYLISTADD s=%s&size=%s&ed2k=%s&edit=%d&viewed=%d&state=1&tag=mlaf%d" % (self.session,size,hash,int(edit),int(watched),i)
        if self.session:
            self.enqueue(msg)
            return True
        else:
            return False
